/**
 * @brief Ashikhmin Tone Mapping Operator
 * 
 * Ashikhmin tone reproduction for displaying high contrast
 * scenes. 
 * Michael Ashikhmin 2002
 *
 * 
 * @author Akiko Yoshida, <yoshida@mpi-sb.mpg.de>
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: pfstmo_ashikhmin02.cpp,v 1.5 2004/12/15 10:11:03 krawczyk Exp $
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pfs.h>
#include <tmo_ashikhmin02.h>
#include <QFile>

using namespace std;

void calculateLuminance( pfs::Array2D* Y, float& avLum, float& maxLum, float& minLum);

#ifdef _WIN32
QByteArray pfstmo_ashikhmin02(const char* input, const char* output,  bool _simple, float _lc_value, int _eq) {
	FILE *inputfp=fopen(input,"rb");
	FILE *fpout=fopen(output,"wb");
#else
QByteArray pfstmo_ashikhmin02(QByteArray inqba,  bool _simple, float _lc_value, int _eq) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
	//prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
	char *pfshdr;
	size_t size;
	FILE * fpout=open_memstream (&pfshdr, &size);
#endif

	pfs::DOMIO pfsio;
	//--- default tone mapping parameters;
	bool simple_flag=_simple;
	float lc_value = _lc_value;
	int eq = _eq;

	pfs::Frame *frame = pfsio.readFrame( inputfp );
	assert(frame != NULL );

	pfs::Channel *X, *Y, *Z;
	frame->getXYZChannels( X, Y, Z );
	assert( X!=NULL || Y!=NULL || Z!=NULL );
	
	float maxLum,avLum, minLum;
	calculateLuminance( Y, avLum, maxLum, minLum);
	
	int w = Y->getCols();
	int h = Y->getRows();
	
	pfs::Array2D* L = new pfs::Array2DImpl(w,h);
	tmo_ashikhmin02(Y, L, maxLum, minLum, avLum, simple_flag, lc_value, eq);

	for( int x=0 ; x<w ; x++ )
		for( int y=0 ; y<h ; y++ )
		{
			float scale = (*L)(x,y) / (*Y)(x,y);
			(*Y)(x,y) *= scale;
			(*X)(x,y) *= scale;
			(*Z)(x,y) *= scale;
		}

	delete L;

	pfsio.writeFrame( frame, fpout );
	pfsio.freeFrame( frame ); // would this delete the previous buffer?

	fclose(inputfp);
	fclose(fpout);
	QByteArray toreturn;
#ifndef _WIN32
	toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
// 	free(pfshdr); //since QByteArray::fromRawData is shallow copy, I cannot free it.
#else
	QFile f(output);
	f.open(QIODevice::ReadOnly);
	toreturn=f.readAll();
#endif
	return toreturn;
}

void calculateLuminance( pfs::Array2D* Y, float& avLum, float& maxLum, float& minLum)
{
  avLum = 0.0f;
  maxLum = 0.0f;
  minLum = 0.0f;

  int size = Y->getCols() * Y->getRows();

  for( int i=0 ; i<size; i++ )
  {
    avLum += log( (*Y)(i) + 1e-4 );
    maxLum = ( (*Y)(i) > maxLum ) ? (*Y)(i) : maxLum ;
    minLum = ( (*Y)(i) < minLum ) ? (*Y)(i) : minLum ;
  }
  avLum =exp( avLum/ size);
}

