/**
 * @file pfstmo_fattal02.cpp
 * @brief Tone map XYZ channels using Fattal02 model
 *
 * Gradient Domain High Dynamic Range Compression
 * R. Fattal, D. Lischinski, and M. Werman
 * In ACM Transactions on Graphics, 2002.
 *
 * 
 * This file is a part of PFSTMO package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2003,2004 Grzegorz Krawczyk
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 * 
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: pfstmo_fattal02.cpp,v 1.3 2005/12/15 14:51:11 krawczyk Exp $
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <pfs.h>
#include <QFile>
#include <tmo_fattal02.h>

using namespace std;

#ifdef _WIN32
QByteArray pfstmo_fattal02(const char* input, const char* output, float _opt_alfa,float _opt_beta,float _opt_saturation) {
	FILE *inputfp=fopen(input,"rb");
        FILE *fpout=fopen(output,"wb");
#else
QByteArray pfstmo_fattal02(QByteArray inqba, float _opt_alfa,float _opt_beta,float _opt_saturation) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
	//prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
	char *pfshdr;
	size_t size;
	FILE * fpout=open_memstream (&pfshdr, &size);
#endif
	
	pfs::DOMIO pfsio;
	//--- default tone mapping parameters;
	float opt_alfa = _opt_alfa;
	float opt_beta = _opt_beta;
	float opt_saturation=_opt_saturation;
	
	pfs::Frame *frame = pfsio.readFrame( inputfp );
	assert( frame != NULL );
	
	pfs::Channel *X, *Y, *Z;
	frame->getXYZChannels( X, Y, Z );
	
	assert( Y!=NULL || X!=NULL || Z!=NULL);
	
	// tone mapping
	int w = Y->getCols();
	int h = Y->getRows();
	
	pfs::Array2D* L = new pfs::Array2DImpl(w,h);
	tmo_fattal02(Y, L, opt_alfa, opt_beta);
	
	for( int x=0 ; x<w ; x++ )
		for( int y=0 ; y<h ; y++ )
		{
			(*X)(x,y) = powf( (*X)(x,y)/(*Y)(x,y), opt_saturation ) * (*L)(x,y);
			(*Z)(x,y) = powf( (*Z)(x,y)/(*Y)(x,y), opt_saturation ) * (*L)(x,y);
			(*Y)(x,y) = (*L)(x,y);
		}
	
	delete L;
	
	pfsio.writeFrame( frame, fpout );
	pfsio.freeFrame( frame );

	fclose(inputfp);
	fclose(fpout);
	QByteArray toreturn;
#ifndef _WIN32
	toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
// 	free(pfshdr); //since QByteArray::fromRawData is shallow copy, I cannot free it.
#else
	QFile f(output);
	f.open(QIODevice::ReadOnly);
	toreturn=f.readAll();
#endif
	return toreturn;
}
